program Gomoku;

{ Designed by Borland International, 1985
  Modified by Erich H. Hauser, Baku, 1999 }

uses Crt;

const
  Ext          =  #0;
  N            =  19;
  CtrlC        =  #3;
  Return       = #13;
  Space        = #32;
  AttackFactor =   4;
  Weight       : array[0..6] of integer = (0, 0, 4, 20, 100, 500, 0);
  PlayerColor  : integer = 11;
  ProgColor    : integer = 10;
  NormalColor  : integer = 15;
  BorderColor  : integer = 14;
  BoardColor   : integer = 7;
  HeadingColor : integer = 12;

type
  TypeOfWin  = (Null, Horiz, DownLeft, DownRight, Vert);
  BoardType  = (Empty, Cross, Nought);
  ColorType  = Cross..Nought;
  IndexType  = 1..N;
  NumberType = 0..5;
  LineType   = array[ColorType] of NumberType;
  ValueType  = array[ColorType] of integer;
  MaxString  = string[255];

var
  Board      : array[IndexType, IndexType] of BoardType;
  Player     : ColorType;
  TotalLines : integer;
  GameWon    : boolean;
  Line       : array[0..3, IndexType, IndexType] of LineType;
  Value      : array[IndexType, IndexType] of ValueType;
  X, Y       : IndexType;
  Command    : char;
  AutoPlay   : boolean;

procedure Abort;
begin
  NormVideo;
  Window(1, 1, 80, 25);
  GotoXY(1, 24);
  ClrScr;
  Halt;
end;

procedure SetUpScreen;
type
  Str5=string[5];

    procedure WriteBoard(N : integer; Top, Middle, Bottom : Str5);
    var I, J : IndexType;

        procedure WriteLetters;
        var I:integer;
        begin
           TextColor(BorderColor);
           Write('  ');
           for I := 1 to N do Write(Chr(Ord('A') + I - 1):2);
           WriteLn;
        end;

        procedure WriteBoardLine(J : integer; S : Str5);
        var I:integer;
        begin
           TextColor(BorderColor);  Write(J:2, ' ');
           TextColor(BoardColor);   Write(s[1]);
           for I := 2 to N - 1 do  Write(S[2], S[3]);
           Write(S[4], S[5]);
           TextColor(BorderColor);
           WriteLn(' ', J:2);
        end;

    begin
        GotoXY(1, 1);
        WriteLetters;
        WriteBoardLine(N, Top);
        for J := N - 1 downto 2 do WriteBoardLine(J, Middle);
        WriteBoardLine(1, Bottom);
        WriteLetters;
    end;

begin
    WriteBoard(N, 'Ŀ',
                  'Ĵ',
                  '');
    TextColor(NormalColor);
end;

procedure GotoSquare(X, Y : IndexType);
begin
    GotoXY(2 + X * 2, N + 2 - Y);
end;

procedure PrintMove(Piece : ColorType; X, Y : IndexType);
const
    PieceChar  : array[ColorType] of char = ('X', '0');
    PieceColor : array[ColorType] of byte = (11, 10);
begin
    TextColor(PieceColor[Piece]);
    GotoXY(49, 9); Write(PieceChar[Piece], Chr(Ord('A') + X - 1):2, Y);
    ClrEOL;
    GotoSquare(X, Y); Write(PieceChar[Piece]);  GotoSquare(X, Y);
    TextColor(NormalColor);
end;

procedure ClearMove;
begin
    GotoXY(49, 9); ClrEOL;
end;

procedure PrintMsg(Str : MaxString);
begin
    TextColor(NormalColor);
    GotoXY(1, 23); Write(Str);
end;

procedure ClearMsg;
begin
    GotoXY(1,23); ClrEOL;
end;

procedure WriteCommand(S : MaxString);
begin
    TextColor(NormalColor); Write(S[1]);
    TextColor(BoardColor); Write(Copy(S, 2, Length(s) - 1));
end;

procedure ResetGame(FirstGame : boolean);
var
    I, J : IndexType;
    D    : 0..3;
    C    : ColorType;
begin
    SetUpScreen;
    if FirstGame then
       begin
           TextColor(HeadingColor);
           GotoXY(49, 1);
           Write('T U R B O - G O M O K U');
           GotoXY(49, 3);
           WriteCommand('Newgame ');
           WriteCommand('Quit ');
           WriteCommand('Autoplay ');
           FirstGame := false;
       end
    else
       begin
           ClearMsg; ClearMove;
       end;
    for I := 1 to N do
    for J := 1 to N do
        begin
            Board[I, J] := Empty;
            for C := Cross to Nought do
                begin
                    Value[I, J, C] := 0;
                    for D := 0 to 3 do Line[D, I, J, C] := 0;
                end;
         end;
         Player := Cross;
         TotalLines := 2 * 2 * (N * (N - 4) + (N - 4) * (N - 4));
         GameWon := false;
end;

function OpponentColor(Player : ColorType) : ColorType;
begin
    if Player = Cross then
       OpponentColor := Nought
    else
       OpponentColor := Cross;
end;

procedure BlinkWinner(Piece : ColorType;
                       X, Y : IndexType;
                WinningLine : TypeOfWin);
const
  PieceChar  : array[ColorType] of char = ('X', '0');
  PieceColor : array[ColorType] of byte = (12, 12);

var
  XHold, YHold : integer;
  Dx, Dy       : integer;

    procedure BlinkRow(X, Y, Dx, Dy : integer);
    var I : integer;
    begin
        TextColor(PieceColor[Piece]);
        for I := 1 to 5 do
            begin
                GotoSquare(X, Y);
                Write(PieceChar[Piece]);
                X := X - Dx;
                Y := Y - Dy;
            end;
    end;

begin
    TextColor(PieceColor[Piece]);
    GotoXY(49, 9); Write(PieceChar[Piece], Chr(Ord('A') + X - 1):2, Y); ClrEOL;
    XHold := X; YHold := Y;
    case WinningLine of
         Horiz :     begin  Dx:=1;  Dy:=0;  end;
         DownLeft :  begin  Dx:=1;  Dy:=1;  end;
         Vert :      begin  Dx:=0;  Dy:=1;  end;
         DownRight : begin  Dx:=-1; Dy:=1;  end;
    end;
    while (Board[X+Dx,Y+Dy]<>Empty) and (Board[X+Dx,Y+Dy]=Piece) do
         begin X:=X+Dx; Y:=Y+Dy; end;
    BlinkRow(X, Y, Dx, Dy); X:=XHold; Y:=YHold;
    GotoSquare(X, Y); TextColor(PlayerColor);
end;

procedure MakeMove(X, Y : IndexType);

var
  Opponent : ColorType;
  X1 ,Y1   : integer;
  K, L     : NumberType;
  WinningLine : TypeOfWin;

    procedure Add(var Num : NumberType);
    begin
        Num := Num + 1;
        if Num = 1 then TotalLines := TotalLines - 1;
        if Num = 5 then GameWon := true;
    end;

    procedure Update(Lin : LineType; var Valu : ValueType);
    begin
        if Lin[Opponent] = 0 then
           Valu[Player] := Valu[Player] +
                           Weight[Lin[Player] + 1] - Weight[Lin[Player]]
        else
           if Lin[Player] = 1 then
              Valu[Opponent] := Valu[Opponent] - Weight[Lin[Opponent] + 1];
    end;

begin
  WinningLine := Null;
  Opponent := OpponentColor(Player);
  GameWon := false;

  for K := 0 to 4 do
  begin
    X1 := X - K;
    Y1 := Y;
    if (1 <= X1) and (X1 <= N - 4) then
    begin
      Add(Line[0, X1, Y1, Player]);
      if GameWon and (WinningLine = Null) then
        WinningLine := Horiz;
      for L := 0 to 4 do
        Update(Line[0, X1, Y1], Value[X1 + L, Y1]);
    end;
  end;

  for K := 0 to 4 do
  begin
    X1 := X - K;
    Y1 := Y - K;
    if (1 <= X1) and (X1 <= N - 4) and
       (1 <= Y1) and (Y1 <= N - 4) then
    begin
      Add(Line[1, X1, Y1, Player]);
      if GameWon and (WinningLine = Null) then
        WinningLine := DownLeft;
      for L := 0 to 4 do
        Update(Line[1, X1, Y1], Value[X1 + L, Y1 + L]);
    end;
  end;

  for K := 0 to 4 do
  begin
    X1 := X + K;
    Y1 := Y - K;
    if (5 <= X1) and (X1 <= N) and
       (1 <= Y1) and (Y1 <= N - 4) then
    begin
      Add(Line[3, X1, Y1, Player]);
      if GameWon and (WinningLine = Null) then
        WinningLine := DownRight;
      for L := 0 to 4 do
        Update(Line[3, X1, Y1], Value[X1 - L, Y1 + L]);
    end;
  end;

  for K := 0 to 4 do
  begin
    X1 := X;
    Y1 := Y - K;
    if (1 <= Y1) and (Y1 <= N - 4) then
    begin
      Add(Line[2, X1, Y1, Player]);
      if GameWon and (WinningLine = Null) then
        WinningLine := Vert;
      for L := 0 to 4 do
        Update(Line[2, X1, Y1], Value[X1, Y1 + L]);
    end;
  end;

  Board[X, Y] := Player;
  if GameWon then
    BlinkWinner(Player, X, Y, WinningLine)
  else
    PrintMove(Player, X, Y);
  Player := Opponent;
end;

function GameOver : boolean;
begin
  GameOver := GameWon or (TotalLines <= 0);
end;

procedure FindMove(var X, Y : IndexType);
var
  Opponent  : ColorType;
  I, J      : IndexType;
  Max, Valu : integer;
begin
  Opponent := OpponentColor(Player);
  Max := -MaxInt;
  X := (N + 1) DIV 2;
  Y := (N + 1) DIV 2;
  if Board[X, Y] = Empty then Max := 4;

   for I := 1 to N do
   for J := 1 to N do
       if Board[I, J] = Empty then
       begin
         Valu := Value[I, J, Player] * (16 + AttackFactor) DIV
                 16 + Value[I, J, Opponent] + Random(4);
         if Valu > Max then begin X := I; Y := J; Max := Valu; end;
       end;
end;

procedure ClearBuffer;
var
  Ch : char;
begin
  While KeyPressed do
    Ch:=ReadKey;
end;

procedure GetChar(var Ch : char);
begin
  Ch:=ReadKey;
  Ch := UpCase(Ch);
end;

procedure ReadCommand(X, Y : IndexType; var Command : char);
var
  ValidCommand : boolean;

begin
  repeat
    ValidCommand := true;
    GotoSquare(X, Y);
    GetChar(Command);
    case Command of
      CtrlC    : Command := 'Q';
      Return,
      Space    : Command := 'E';
      Ext      : begin
                   if KeyPressed then
                   begin
                     Command:=ReadKey;
                     case Command of
                       'K' : Command := 'L';
                       'M' : Command := 'R';
                       'P' : Command := 'D';
                       'H' : Command := 'U';
                       'G' : Command := '7';
                       'I' : Command := '9';
                       'O' : Command := '1';
                       'Q' : Command := '3';
                       else
                       begin
                         ValidCommand := false;
                         ClearBuffer;
                       end;
                     end;
                   end
                   else
                     if GameOver then command := 'P'
                     else
                     begin
                       ValidCommand := false;
                       ClearBuffer;
                     end;
                 end;
      'N','Q','A','P','H' : ;
      else
      begin
        ValidCommand := false;
        ClearBuffer;
      end;
    end;
  until ValidCommand;
end;

procedure Initialize;
begin
  ClrScr;
  Randomize;
  AutoPlay := false;
end;

procedure InterpretCommand(Command : char);
var
  Temp : integer;
begin
  case Command of
    'N': begin
           ResetGame(false);
           X := (N + 1) DIV 2;
           Y := X;
         end;
    'L': X := (X + N - 2) MOD N + 1;
    'R': X := X MOD N + 1;
    'D': Y := (Y + N - 2) MOD N + 1;
    'U': Y := Y MOD N + 1;
    '7': begin
           if (X = 1) or (Y = N) then
           begin
             Temp := X;
             X := Y;
             Y := Temp;
           end
           else
           begin
             X := X - 1;
             Y := Y + 1;
           end;
         end;
    '9': begin
           if X = N then
           begin
             X := (N - Y) + 1;
             Y := 1;
           end
           else if Y = N then
           begin
             Y := (N - X) + 1;
             X := 1;
           end
           else
           begin
             X := X + 1;
             Y := Y + 1;
           end
         end;
    '1': begin
           if Y = 1 then
           begin
             Y := (N - X) + 1;
             X := N;
           end
           else if X = 1 then
           begin
             X := (N - Y) + 1;
             Y := N;
           end
           else
           begin
             X := X - 1;
             Y := Y - 1;
           end;
         end;
    '3': begin
           if (X = N) or (Y = 1) then
           begin
             Temp := X;
             X := Y;
             Y := Temp;
           end
           else
           begin
             X := X + 1;
             Y := Y - 1;
           end;
         end;
    'A': AutoPlay := true;
  end;
end;

procedure PlayerMove;
begin
  if Board[X, Y] = Empty then
  begin
    MakeMove(X, Y);
    if GameWon then
      PrintMsg('Congratulations, You won!');
    Command := 'P';
  end;
end;

procedure ProgramMove;
begin
  repeat
    if KeyPressed then
      ClearBuffer;
    if GameOver then
    begin
      AutoPlay := false;
      if (Command <> 'Q') and (not GameWon) then
        PrintMsg('Tie game!');
    end
    else
    begin
      FindMove(X, Y);
      MakeMove(X, Y);
      if GameWon then
        PrintMsg('I won!');
    end;
  until AutoPlay = false;
end;

begin
  Initialize;
  ResetGame(true);
  X := (N + 1) DIV 2;
  Y := X;
  repeat
    ReadCommand(X, Y, Command);
    if GameOver then if Command <> 'Q' then Command := 'N';
    InterpretCommand(Command);
    if Command = 'E' then begin PlayerMove; ProgramMove; end;
    if Command = 'A' then ProgramMove;
  until Command in ['Q', CtrlC];
  Abort;
end.
